"""
Module to provide helpers for the loaders to share.
"""

# pylint: disable=too-few-public-methods
from typing import Callable, Optional


class ApplicationPropertiesLoaderHelper:
    """
    Class to provide helpers for the loaders to share.
    """

    @staticmethod
    def set_error_handler_if_not_set(
        handle_error_fn: Optional[Callable[[str, Optional[Exception]], None]],
    ) -> Callable[[str, Optional[Exception]], None]:
        """
        If the error_handler is not set, make sure to set it to print to output.
        """
        if not handle_error_fn:

            def print_error_to_stdout(
                formatted_error: str, thrown_exception: Optional[Exception]
            ) -> None:
                _ = thrown_exception
                print(formatted_error)

            handle_error_fn = print_error_to_stdout
        return handle_error_fn


# pylint: enable=too-few-public-methods



"""
Module to provide for easy access to various property file formats.

https://stackoverflow.com/questions/44834/what-does-all-mean-in-python#When%20Avoiding%20__all__%20Makes%20Sense
"""

from application_properties.application_properties import (  # noqa F401
    ApplicationProperties,
)
from application_properties.application_properties_config_loader import (  # noqa F401
    ApplicationPropertiesConfigLoader,
)
from application_properties.application_properties_facade import (  # noqa F401
    ApplicationPropertiesFacade,
)
from application_properties.application_properties_json_loader import (  # noqa F401
    ApplicationPropertiesJsonLoader,
)
from application_properties.application_properties_toml_loader import (  # noqa F401
    ApplicationPropertiesTomlLoader,
)
from application_properties.application_properties_utilities import (  # noqa F401
    ApplicationPropertiesUtilities,
)
from application_properties.application_properties_yaml_loader import (  # noqa F401
    ApplicationPropertiesYamlLoader,
)

__all__ = [
    "ApplicationProperties",
    "ApplicationPropertiesUtilities",
    "ApplicationPropertiesFacade",
    "ApplicationPropertiesJsonLoader",
    "ApplicationPropertiesTomlLoader",
    "ApplicationPropertiesYamlLoader",
    "ApplicationPropertiesConfigLoader",
]



"""
Library version information.
"""

__version__ = "0.9.2"
__project_name__ = "application_properties"
__description__ = (
    "A simple, easy to use, unified manner of accessing program properties."
)



"""
Module to provide for a facade in front of an ApplicationProperties instance that
only exposes part of the properties tree.
"""

import logging
from typing import Any, Callable, List, Optional

from application_properties.application_properties import ApplicationProperties

LOGGER = logging.getLogger(__name__)


class ApplicationPropertiesFacade:
    """
    Class to provide for a facade in front of an ApplicationProperties instance that
    only exposes part of the properties tree.
    """

    def __init__(
        self, base_properties: ApplicationProperties, property_prefix: str
    ) -> None:
        """
        Initializes an new instance of the ApplicationPropertiesFacade class.
        """
        if not isinstance(base_properties, ApplicationProperties):
            raise ValueError(
                "The base_properties of the facade must be an ApplicationProperties instance."
            )
        self.__base_properties = base_properties

        if not isinstance(property_prefix, str):
            raise ValueError("The property_prefix argument must be a string.")
        if not property_prefix.endswith(base_properties.separator):
            raise ValueError(
                f"The property_prefix argument must end with the separator character '{base_properties.separator}'."
            )
        self.__property_prefix = property_prefix

    # pylint: disable=too-many-arguments
    def get_property(
        self,
        property_name: str,
        property_type: type,
        default_value: Any = None,
        valid_value_fn: Optional[Callable[[Any], Any]] = None,
        is_required: bool = False,
        strict_mode: Optional[Any] = None,
    ) -> Any:
        """
        Get an property of a generic type from the configuration.
        """

        return self.__base_properties.get_property(
            f"{self.__property_prefix}{property_name}",
            property_type,
            default_value=default_value,
            valid_value_fn=valid_value_fn,
            is_required=is_required,
            strict_mode=strict_mode,
        )

    # pylint: enable=too-many-arguments

    def get_boolean_property(
        self,
        property_name: str,
        default_value: Optional[bool] = None,
        is_required: bool = False,
    ) -> Optional[bool]:
        """
        Get a boolean property from the configuration.
        """
        return self.__base_properties.get_boolean_property(
            f"{self.__property_prefix}{property_name}",
            default_value=default_value,
            is_required=is_required,
        )

    # pylint: disable=too-many-arguments
    def get_integer_property(
        self,
        property_name: str,
        default_value: Optional[int] = None,
        valid_value_fn: Optional[Callable[[int], Any]] = None,
        is_required: bool = False,
        strict_mode: Optional[bool] = None,
    ) -> Optional[int]:
        """
        Get an integer property from the configuration.
        """
        return self.__base_properties.get_integer_property(
            f"{self.__property_prefix}{property_name}",
            default_value=default_value,
            valid_value_fn=valid_value_fn,
            is_required=is_required,
            strict_mode=strict_mode,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def get_string_property(
        self,
        property_name: str,
        default_value: Optional[str] = None,
        valid_value_fn: Optional[Callable[[str], Any]] = None,
        is_required: bool = False,
        strict_mode: Optional[bool] = None,
    ) -> Optional[str]:
        """
        Get a string property from the configuration.
        """
        return self.__base_properties.get_string_property(
            f"{self.__property_prefix}{property_name}",
            default_value=default_value,
            valid_value_fn=valid_value_fn,
            is_required=is_required,
            strict_mode=strict_mode,
        )

    # pylint: enable=too-many-arguments

    @property
    def property_names(self) -> List[str]:
        """
        List of each of the properties in the map.
        """
        return [
            next_property_name[len(self.__property_prefix) :]
            for next_property_name in self.__base_properties.property_names
            if next_property_name.startswith(self.__property_prefix)
        ]

    def property_names_under(self, key_name: str) -> List[str]:
        """
        List of each of the properties in the map under the specified key.
        """
        ApplicationProperties.verify_full_key_form(key_name)
        return [
            next_key_name
            for next_key_name in self.property_names
            if next_key_name.startswith(key_name)
        ]



"""
Module containing the MultisourceConfigurationLoader class and its supporting classes.
"""

import contextlib
import json
import logging
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Callable, List, Optional, Tuple

import pyjson5
import tomli
import yaml

from application_properties import ApplicationProperties
from application_properties.application_properties_json_loader import (
    ApplicationPropertiesJsonLoader,
)
from application_properties.application_properties_loader_helper import (
    ApplicationPropertiesLoaderHelper,
)
from application_properties.application_properties_toml_loader import (
    ApplicationPropertiesTomlLoader,
)
from application_properties.application_properties_yaml_loader import (
    ApplicationPropertiesYamlLoader,
)

LOGGER = logging.getLogger(__name__)


class ConfigurationFileType(Enum):
    """Represents the supported configuration file types for application properties.

    Used to distinguish between JSON, YAML, and TOML file types.
    """

    NONE = 0
    JSON = 1
    YAML = 2
    YML = 3
    TOML = 4


@dataclass
class MultisourceConfigurationLoaderOptions:
    """
    Options to modify the behavior of one of more configuration sources
    during the MultisourceConfigurationLoader processing of registered sources.
    """

    load_json_files_as_json5: bool = False
    """
    Allow parsing of JSON files using the JSON5 parser. (Default = False)
    """
    section_header_if_toml: Optional[str] = None
    """Optional section header to use when loading TOML configuration files.
    """


# pylint: disable=too-few-public-methods
class BaseConfigurationSource(ABC):
    """
    Base configuration source.
    """

    _file_type_to_extension_map = {
        ConfigurationFileType.JSON: ".json",
        ConfigurationFileType.YAML: ".yaml",
        ConfigurationFileType.YML: ".yml",
        ConfigurationFileType.TOML: ".toml",
    }

    @abstractmethod
    def apply_configuration(
        self,
        options: MultisourceConfigurationLoaderOptions,
        application_properties: ApplicationProperties,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> Tuple[bool, bool]:
        """
        Apply the current configuration source's configuration to the specified `ApplicationProperties` object.

        Args:
            options: Options to change the behavior of one of more configuration sources.
            application_properties: Instance of `ApplicationProperties` to apply the configuration to.
            handle_error_fn: Function to call if there are any errors when applying the configuration.
        """

    def __load_as_json(
        self,
        file_name: str,
        options: MultisourceConfigurationLoaderOptions,
        application_properties: ApplicationProperties,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> Tuple[bool, bool]:
        LOGGER.debug(
            "Attempting to find/load '%s' as a JSON configuration file.",
            file_name,
        )
        (
            did_apply_map,
            did_have_one_error,
        ) = ApplicationPropertiesJsonLoader.load_and_set(
            application_properties,
            file_name,
            handle_error_fn=handle_error_fn,
            clear_property_map=False,
            check_for_file_presence=True,
            load_as_json5_file=options.load_json_files_as_json5,
        )
        return did_apply_map, did_have_one_error

    def __load_as_yaml(
        self,
        file_name: str,
        application_properties: ApplicationProperties,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> Tuple[bool, bool]:

        LOGGER.debug(
            "Attempting to find/load '%s' as a YAML configuration file.",
            file_name,
        )
        (
            did_apply_map,
            did_have_one_error,
        ) = ApplicationPropertiesYamlLoader.load_and_set(
            application_properties,
            file_name,
            handle_error_fn=handle_error_fn,
            clear_property_map=False,
            check_for_file_presence=True,
        )
        return did_apply_map, did_have_one_error

    def __load_as_toml(
        self,
        file_name: str,
        options: MultisourceConfigurationLoaderOptions,
        application_properties: ApplicationProperties,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> Tuple[bool, bool]:

        LOGGER.debug(
            "Attempting to find/load '%s' as a TOML configuration file.",
            file_name,
        )
        (
            did_apply_map,
            did_have_one_error,
        ) = ApplicationPropertiesTomlLoader.load_and_set(
            application_properties,
            file_name,
            section_header=options.section_header_if_toml,
            handle_error_fn=handle_error_fn,
            clear_property_map=False,
            check_for_file_presence=True,
        )
        return did_apply_map, did_have_one_error

    # pylint: disable=too-many-arguments
    def _load_config(
        self,
        config_file_type: ConfigurationFileType,
        file_name: str,
        options: MultisourceConfigurationLoaderOptions,
        application_properties: ApplicationProperties,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> Tuple[bool, bool]:

        if config_file_type == ConfigurationFileType.JSON:
            return self.__load_as_json(
                file_name, options, application_properties, handle_error_fn
            )
        if config_file_type in [
            ConfigurationFileType.YAML,
            ConfigurationFileType.YML,
        ]:
            return self.__load_as_yaml(
                file_name, application_properties, handle_error_fn
            )
        assert config_file_type == ConfigurationFileType.TOML
        return self.__load_as_toml(
            file_name, options, application_properties, handle_error_fn
        )
        # return False, False

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class LocalPyprojectTomlFile(BaseConfigurationSource):
    """
    Class to allow the multisource configuration loader to reference a
    "pyproject.toml" file that may or may not exist.
    """

    __pyproject_toml_file = "pyproject.toml"

    def __init__(self, section_header_name: str) -> None:
        """
        Create an instance of the LocalPyprojectTomlFile object.

        Args:
            section_header_name: Name of the section within the `pyproject.toml`
                file containing the applications configuration.
        """
        self.__section_header_name = section_header_name

    def apply_configuration(
        self,
        options: MultisourceConfigurationLoaderOptions,
        application_properties: ApplicationProperties,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> Tuple[bool, bool]:

        _ = options

        LOGGER.debug(
            "Looking for local '%s' file.", LocalPyprojectTomlFile.__pyproject_toml_file
        )
        project_configuration_file = os.path.abspath(
            LocalPyprojectTomlFile.__pyproject_toml_file
        )
        return ApplicationPropertiesTomlLoader.load_and_set(
            application_properties,
            project_configuration_file,
            self.__section_header_name,
            handle_error_fn,
            clear_property_map=False,
            check_for_file_presence=True,
        )


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class LocalProjectConfigurationFile(BaseConfigurationSource):
    """
    Class to allow the multisource configuration loader to reference a
    project configuration file that may or may not exist.
    """

    def __init__(
        self,
        project_file_name: str,
        config_file_type: ConfigurationFileType,
        alternate_extension_types: Optional[List[ConfigurationFileType]] = None,
    ) -> None:
        """
        Create an instance of the LocalPyprojectTomlFile object.

        When the `apply_configuration` function is called, it will first
        try and load the specified `project_file_name` file using the
        file type specified by `config_file_type`.  If a file with that name
        does not exist, it will change the extension of the file specified
        in `project_file_name` to zero or more of the extensions provided
        in the `alternate_extension_types` argument.  Note that only if
        the file is not present with that exact name will it proceed with
        using other extensions.

        Args:
            project_file_name: Name of the project configuration file to load.
            config_file_type: Format of data within the configuration file.
            alternate_extension_types: Optional set of file types to use by
                looking for project files with the same basename but different
                extensions.
        Raises:
            ValueError: If the config_file_type parameter is `ConfigurationFileType.NONE`.
        """
        self.project_file_name = project_file_name
        self.config_file_type = config_file_type
        self.alternate_extension_types = alternate_extension_types

        if self.config_file_type == ConfigurationFileType.NONE:
            raise ValueError(
                "Project configuration file must have a non-NONE file type set."
            )

    def apply_configuration(
        self,
        options: MultisourceConfigurationLoaderOptions,
        application_properties: ApplicationProperties,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> Tuple[bool, bool]:

        LOGGER.debug("Looking for local configuration files.")

        base_file_name = self.project_file_name
        abs_file_name = os.path.abspath(self.project_file_name)

        expected_file_type = BaseConfigurationSource._file_type_to_extension_map[
            self.config_file_type
        ]
        if base_file_name.endswith(expected_file_type):
            base_file_name = base_file_name[: -len(expected_file_type)]

        did_apply_map, did_have_one_error = self._load_config(
            self.config_file_type,
            abs_file_name,
            options,
            application_properties,
            handle_error_fn,
        )

        while (
            self.alternate_extension_types
            and not did_apply_map
            and not did_have_one_error
        ):
            top_alternate_extension_type = self.alternate_extension_types.pop(0)
            new_file_name = (
                os.path.abspath(base_file_name)
                + BaseConfigurationSource._file_type_to_extension_map[
                    top_alternate_extension_type
                ]
            )
            did_apply_map, did_have_one_error = self._load_config(
                top_alternate_extension_type,
                new_file_name,
                options,
                application_properties,
                handle_error_fn,
            )

        if not did_apply_map:
            LOGGER.debug("No default configuration files were loaded.")
        return did_apply_map, did_have_one_error


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class SpecifiedConfigurationFile(BaseConfigurationSource):
    """
    Class to allow the multisource configuration loader to reference a
    specific configuration file provided, typically by the user.
    """

    def __init__(
        self,
        specified_file_name: Optional[str],
        config_file_type: ConfigurationFileType,
    ) -> None:
        """
        Create an instance of the SpecifiedConfigurationFile object.

        During the invocation of the `apply_configuration` function for this class,
        if the `config_file_type` parameter is `ConfigurationFileType.NONE`, a first
        pass will check to see if the extension of the file matches any of the file
        extensions registered to the configuration file types, using that as the
        configuration file type if a match is made. If that check fails to determine
        a file type, the second pass will check the contents of the file by using the
        base loaders for each of the file types.

        Args:
            specified_file_name: Name of configuration file to try and load.  If the name
                                 of the configuration file is empty or None, no attempt
                                 is made to load the file.
            config_file_type: Type of the configuration file being loaded.
        """
        self.specified_file_name = specified_file_name
        self.config_file_type = config_file_type

    # pylint: disable=no-member
    def __determine_file_type(
        self, options: MultisourceConfigurationLoaderOptions
    ) -> ConfigurationFileType:

        file_type = self.config_file_type
        assert self.specified_file_name

        if file_type == ConfigurationFileType.NONE:
            for i in ConfigurationFileType:
                if (
                    i != ConfigurationFileType.NONE
                    and self.specified_file_name.endswith(
                        BaseConfigurationSource._file_type_to_extension_map[i]
                    )
                ):
                    file_type = i
                    break
        if file_type == ConfigurationFileType.NONE:
            LOGGER.debug(
                "Attempt to determine file type for specified configuration file '%s' by content.",
                self.specified_file_name,
            )
            with open(self.specified_file_name, encoding="utf-8") as infile:
                if options.load_json_files_as_json5:
                    with contextlib.suppress(pyjson5.Json5DecoderException):
                        pyjson5.load(infile)
                        file_type = ConfigurationFileType.JSON
                else:
                    with contextlib.suppress(json.decoder.JSONDecodeError):
                        json.load(infile)
                        file_type = ConfigurationFileType.JSON
        if file_type == ConfigurationFileType.NONE:
            with contextlib.suppress(yaml.MarkedYAMLError):
                with open(self.specified_file_name, "rb") as infile:
                    _ = yaml.safe_load(infile)
                file_type = ConfigurationFileType.YAML
        if file_type == ConfigurationFileType.NONE:
            with contextlib.suppress(tomli.TOMLDecodeError):
                with open(self.specified_file_name, "rb") as infile:
                    tomli.load(infile)
                file_type = ConfigurationFileType.TOML
        return file_type

    # pylint: enable=no-member

    def apply_configuration(
        self,
        options: MultisourceConfigurationLoaderOptions,
        application_properties: ApplicationProperties,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> Tuple[bool, bool]:

        if not self.specified_file_name:
            return False, False

        if not os.path.isfile(self.specified_file_name):
            handle_error_fn(
                f"Specified configuration file `{self.specified_file_name}` does not exist.",
                None,
            )
            return False, True

        file_type = self.__determine_file_type(options)
        if file_type == ConfigurationFileType.NONE:
            formatted_error = f"Specified configuration file '{self.specified_file_name}' was not parseable as a JSON, YAML, or TOML file."
            LOGGER.warning(formatted_error)
            handle_error_fn(formatted_error, None)
            return False, True

        did_apply_map, did_have_one_error = self._load_config(
            file_type,
            self.specified_file_name,
            options,
            application_properties,
            handle_error_fn,
        )
        return did_apply_map, did_have_one_error


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class ManuallySetProperties(BaseConfigurationSource):
    """
    Class to allow the multisource configuration loader to reference a
    zero or more individual properties, typically provided by the user.
    """

    def __init__(self, manual_properties: Optional[List[str]]) -> None:
        """
        Create an instance of the ManuallySetProperties object.

        During the invocation of the `apply_configuration` function for this class,
        if any string is not correctly formatted, the `handle_error_fn` function
        will be invoked stating the specific string that is not correctly formatted.

        For more information on the format of the property strings,
        consult the [documentation page](https://github.com/jackdewinter/application_properties/blob/main/docs/examples.md).

        Args:
            manual_properties: Optional list of strings properly formatted
                to denote the property keys and the values to set to them.
        """
        self.manual_properties = manual_properties

    def apply_configuration(
        self,
        options: MultisourceConfigurationLoaderOptions,
        application_properties: ApplicationProperties,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> Tuple[bool, bool]:

        _ = options

        did_apply = did_error = False
        if self.manual_properties:
            last_next_item = ""
            try:
                for next_item in self.manual_properties:
                    last_next_item = next_item
                    LOGGER.debug(
                        "Attempting to set manual property '%s'.",
                        next_item,
                    )
                    application_properties.set_manual_property(next_item)
                did_apply = True
            except ValueError as this_exception:
                formatted_error = f"Manually set property '{last_next_item}' was not validly formed: {this_exception}"
                handle_error_fn(formatted_error, this_exception)
                did_error = True

        return did_apply, did_error


# pylint: enable=too-few-public-methods


class MultisourceConfigurationLoader:
    """
    Class that serves as a container for classes descended from `BaseConfigurationSource`
    that are used to build the layers for configuration.
    """

    def __init__(
        self, options: Optional[MultisourceConfigurationLoaderOptions] = None
    ) -> None:
        """
        Create an instance of the `MultisourceConfigurationLoader` class with which
        to build the layers of configuration to apply and the order in which to
        apply them.

        Args:
            options: Optional settings that affect the processing of the
                configuration sources.
        """
        self.__configuration_sources: List[BaseConfigurationSource] = []
        self.__options = (
            options if options is not None else MultisourceConfigurationLoaderOptions()
        )

    def add_local_pyproject_toml_file(
        self, section_header_name: str
    ) -> "MultisourceConfigurationLoader":
        """
        Request that the multisource configuration loader tries to reference
        a "pyproject.toml" file that may or may not exist in the current directory.

        Args:
            section_header: Section header to look for within the "pyproject.toml" file.

        Returns:
            Instance of `self` for chaining `add_*` functions and the `process` function
            in a fluid manner.
        """
        self.__configuration_sources.append(LocalPyprojectTomlFile(section_header_name))
        return self

    def add_local_project_configuration_file(
        self,
        default_file_name: str,
        config_file_type: ConfigurationFileType,
        alternate_extension_types: Optional[List[ConfigurationFileType]] = None,
    ) -> "MultisourceConfigurationLoader":
        """
        Request that the multisource configuration loader tries to reference
        a project based configuration file that may exist in the current directory.

        When the `process` function is called, it will first
        try and load the specified `project_file_name` file using the
        file type specified by `config_file_type`.  If a file with that name
        does not exist, it will change the extension of the file specified
        in `project_file_name` to zero or more of the extensions provided
        in the `alternate_extension_types` argument.  Note that only if
        the file is not present with that exact name will it proceed with
        using other extensions.

        Args:
            default_file_name: Name of the default file to try and load.
            alternate_extension_types
        Returns:
            Instance of `self` for chaining `add_*` functions and the `process` function
            in a fluid manner.
        Raises:
            ValueError: If the config_file_type parameter is `ConfigurationFileType.NONE`.
        """
        self.__configuration_sources.append(
            LocalProjectConfigurationFile(
                default_file_name, config_file_type, alternate_extension_types
            )
        )
        return self

    def add_specified_configuration_file(
        self,
        config_file_name: Optional[str],
        config_file_type: ConfigurationFileType = ConfigurationFileType.NONE,
    ) -> "MultisourceConfigurationLoader":
        """
        Request that the multisource configuration loader tries to reference
        a configuration file that has been explicitly specified by the calling
        application.

        Note:
            During the invocation of the `process` function, if the `config_file_type`
            parameter is `ConfigurationFileType.NONE`, a first
            pass will check to see if the extension of the file matches any of the file
            extensions registered to the configuration file types, using that as the
            configuration file type if a match is made. If that check fails to determine
            a file type, the second pass will check the contents of the file by using the
            base loaders for each of the file types.


        Args:
            config_file_name: Name of configuration file to try and load.  If the name
                              of the configuration file is empty or None, no attempt
                              is made to load the file.
            config_file_type: Type of configuration file to load.  For auto-detecting
                              the file type, see the notes above.
            section_header_if_toml: Optional section header to use when loading
                              TOML configuration files.

        Returns:
            Instance of `self` for chaining `add_*` functions and the `process` function
            in a fluid manner.
        """
        self.__configuration_sources.append(
            SpecifiedConfigurationFile(config_file_name, config_file_type)
        )
        return self

    def add_manually_set_properties(
        self, manual_properties_to_set: Optional[List[str]] = None
    ) -> "MultisourceConfigurationLoader":
        """
        Request that the multisource configuration loader tries to use an
        array of strings to manually set configuration items for the user.
        Note that these strings are most often supplied as arguments to the
        command line of the application.

        During the invocation of the `process` function,
        if any string is not correctly formatted, the `handle_error_fn` function
        will be invoked stating the specific string that is not correctly formatted.

        For more information on the format,
        check [the documentation](https://github.com/jackdewinter/application_properties/blob/main/docs/examples.md)
        here.

        Args:
            manual_properties_to_set: Name of the default file to try and load.
        Returns:
            Instance of `self` for chaining `add_*` functions and the `process` function
            in a fluid manner.
        """
        self.__configuration_sources.append(
            ManuallySetProperties(manual_properties_to_set)
        )
        return self

    def add_custom_source(
        self, source_to_add: BaseConfigurationSource
    ) -> "MultisourceConfigurationLoader":
        """
        Request that the multisource configuration loader uses a custom
        configuration source.

        Args:
            source_to_add: Instance of a class that inherits from the `BaseConfigurationSource`
                class.
        Returns:
            Instance of `self` for chaining `add_*` functions and the `process` function
            in a fluid manner.
        Raises:
            ValueError: If the value provided to `source_to_add` is not an instance
                of the `BaseConfigurationSource` class.
        """
        if not isinstance(source_to_add, BaseConfigurationSource):
            raise ValueError(
                f"Added source '{source_to_add}' is not a valid configuration source."
            )

        self.__configuration_sources.append(source_to_add)
        return self

    def process(
        self,
        application_properties: ApplicationProperties,
        handle_error_fn: Optional[Callable[[str, Optional[Exception]], None]] = None,
    ) -> bool:
        """
        Process any registered configuration sources, stopping at the first sign of
        error.

        Args:
            application_properties: Instance of `ApplicationProperties` to apply the configuration to.
            handle_error_fn: Function to call if there are any errors when applying the configuration.
        """
        guaranteed_handle_error_fn: Callable[[str, Optional[Exception]], None] = (
            ApplicationPropertiesLoaderHelper.set_error_handler_if_not_set(
                handle_error_fn
            )
        )

        for next_configuration_source in self.__configuration_sources:
            _, did_error = next_configuration_source.apply_configuration(
                self.__options, application_properties, guaranteed_handle_error_fn
            )
            if did_error:
                return True
        return False



"""
Module that provides for utilities that help with application_properties usage.
"""

import argparse
import logging
import os
from argparse import ArgumentParser
from typing import Callable, Optional

from application_properties import ApplicationProperties
from application_properties.application_properties_json_loader import (
    ApplicationPropertiesJsonLoader,
)
from application_properties.application_properties_toml_loader import (
    ApplicationPropertiesTomlLoader,
)

LOGGER = logging.getLogger(__name__)


class ApplicationPropertiesUtilities:
    """
    Class that provides for utilities that help with application_properties usage.
    """

    __pyproject_toml_file = "pyproject.toml"
    __pyproject_section_header = "tool.pymarkdown"

    @staticmethod
    def add_default_command_line_arguments(parser: ArgumentParser) -> None:
        """
        Add command line arguments usually used with the application_properties package.
        """

        parser.add_argument(
            "--config",
            "-c",
            dest="configuration_file",
            action="store",
            default=None,
            help="path to the configuration file to use",
        )
        parser.add_argument(
            "--set",
            "-s",
            dest="set_configuration",
            action="append",
            default=None,
            help="manually set an individual configuration property",
            type=ApplicationProperties.verify_manual_property_form,
        )
        parser.add_argument(
            "--strict-config",
            dest="strict_configuration",
            action="store_true",
            default=False,
            help="throw an error if configuration is bad, instead of assuming default",
        )

    @staticmethod
    def process_project_specific_json_configuration(
        default_configuration_file_name: str,
        args: argparse.Namespace,
        application_properties: "ApplicationProperties",
        handle_error_fn: Optional[Callable[[str, Optional[Exception]], None]] = None,
    ) -> None:
        """
        Load configuration information from JSON configuration files.
        """

        # Look for the default configuration file in the current working directory.
        ApplicationPropertiesJsonLoader.load_and_set(
            application_properties,
            os.path.abspath(default_configuration_file_name),
            handle_error_fn,
            clear_property_map=False,
            check_for_file_presence=True,
        )

        # A configuration file specified on the command line has a higher precedence
        # than anything except a specific setting applied on the command line.
        if args.configuration_file:
            ApplicationPropertiesJsonLoader.load_and_set(
                application_properties,
                args.configuration_file,
                handle_error_fn,
                clear_property_map=False,
                check_for_file_presence=False,
            )

        # A specific setting applied on the command line has the highest precedence.
        if args.set_configuration:
            application_properties.set_manual_property(args.set_configuration)

    @staticmethod
    def process_standard_python_configuration_files(
        application_properties: "ApplicationProperties",
        handle_error_fn: Optional[Callable[[str, Optional[Exception]], None]] = None,
    ) -> None:
        """
        Load configuration information from standard Python project files. This currently
        means that application_properties will look for the `[tool.pymarkdown]` section
        within the `pyproject.toml` located within the current directory.
        """

        # Currently, we only support the pyproject.toml file.
        project_configuration_file = os.path.abspath(
            ApplicationPropertiesUtilities.__pyproject_toml_file
        )
        ApplicationPropertiesTomlLoader.load_and_set(
            application_properties,
            project_configuration_file,
            ApplicationPropertiesUtilities.__pyproject_section_header,
            handle_error_fn,
            False,
            True,
        )



"""
Module that provides for an encapsulation of properties for an application.
"""

import contextlib
import copy
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

LOGGER = logging.getLogger(__name__)


class ApplicationProperties:
    """
    Class that provides for an encapsulation of properties for an application.
    """

    __separator = "."
    __assignment_operator = "="
    __manual_property_type_prefix = "$"
    __manual_property_type_string = "$"
    __manual_property_type_integer = "#"
    __manual_property_type_boolean = "!"

    """
    Class to provide for a container of properties that belong to the application.
    """

    def __init__(
        self,
        strict_mode: bool = False,
        convert_untyped_if_possible: bool = False,
        allow_separator_in_keys: bool = False,
    ) -> None:
        """
        Initializes an new instance of the ApplicationProperties class.
        """
        self.__flat_property_map: Dict[str, Any] = {}
        self.__strict_mode: bool = strict_mode
        self.__convert_untyped_if_possible: bool = convert_untyped_if_possible
        self.__allow_separator_in_keys = allow_separator_in_keys

    @property
    def separator(self) -> str:
        """
        Separator used to split the hierarchy of the property names.
        """
        return self.__separator

    @property
    def number_of_properties(self) -> int:
        """
        Number of properties that exist in the map.
        """
        return len(self.property_names)

    @property
    def property_names(self) -> List[str]:
        """
        List of each of the properties in the map.
        """
        return [
            next_item
            for next_item in self.__flat_property_map
            if not next_item.startswith(ApplicationProperties.__separator)
        ]

    @property
    def strict_mode(self) -> bool:
        """
        Gets whether strict mode is on by default.
        """
        return self.__strict_mode

    def enable_strict_mode(self) -> None:
        """
        Sets strict mode to True to enable it.
        """
        self.__strict_mode = True

    @property
    def convert_untyped_if_possible(self) -> bool:
        """
        Gets whether the package is allowed to try and unconvert untyped entries.
        """
        return self.__convert_untyped_if_possible

    def enable_convert_untyped_if_possible(self) -> None:
        """
        Sets convert_untyped_if_possible to True to enable it.
        """
        self.__convert_untyped_if_possible = True

    def clear(self) -> None:
        """
        Clear the configuration map.
        """
        self.__flat_property_map.clear()

    def load_from_dict(
        self,
        config_map: Dict[Any, Any],
        clear_map: bool = True,
        allow_periods_in_keys: bool = False,
    ) -> None:
        """
        Load the properties from a provided dictionary.
        """

        if not isinstance(config_map, dict):
            raise ValueError("Specified parameter was not a dictionary.")

        LOGGER.debug("Loading from dictionary: {%s}", str(config_map))
        if clear_map:
            self.clear()
        self.__scan_map(
            config_map, "", allow_periods_in_keys and self.__allow_separator_in_keys
        )

    @staticmethod
    def verify_full_part_form(property_key: str) -> str:
        """
        Given one part of a full key, verify that it is composed properly.
        """

        if (
            " " in property_key
            or "\t" in property_key
            or "\n" in property_key
            or ApplicationProperties.__assignment_operator in property_key
            or ApplicationProperties.__separator in property_key
        ):
            raise ValueError(
                "Each part of the property key cannot contain a whitespace character, "
                + f"a '{ApplicationProperties.__assignment_operator}' character, or "
                + f"a '{ApplicationProperties.__separator}' character."
            )
        if not property_key:
            raise ValueError(
                "Each part of the property key must contain at least one character."
            )
        return property_key

    @staticmethod
    def verify_full_key_form(
        property_key: str, alternate_name: Optional[str] = None
    ) -> str:
        """
        Given a full key, verify that it is composed properly.
        """

        key_name = alternate_name or "Full property key"

        if property_key.startswith(
            ApplicationProperties.__separator
        ) or property_key.endswith(ApplicationProperties.__separator):
            raise ValueError(
                f"{key_name} must not start or end with the '{ApplicationProperties.__separator}' character."
            )
        doubles = (
            f"{ApplicationProperties.__separator}{ApplicationProperties.__separator}"
        )
        doubles_index = property_key.find(doubles)
        if doubles_index != -1:
            raise ValueError(
                f"{key_name} cannot contain multiples of "
                + f"the {ApplicationProperties.__separator} without any text between them."
            )
        split_key = property_key.split(ApplicationProperties.__separator)
        for next_key in split_key:
            ApplicationProperties.verify_full_part_form(next_key)
        return property_key

    @staticmethod
    def verify_manual_property_form(string_to_verify: str) -> str:
        """
        Verify the general form of a manual property string. i.e. key=value
        """

        if not isinstance(string_to_verify, str):
            raise ValueError("Manual property form must be a string.")
        equals_index = string_to_verify.find(
            ApplicationProperties.__assignment_operator
        )
        if equals_index == -1:
            raise ValueError(
                "Manual property key and value must be separated by the '=' character."
            )
        property_key = string_to_verify[:equals_index]
        ApplicationProperties.verify_full_key_form(property_key)
        return string_to_verify

    def set_manual_property(self, combined_string: Union[str, List[str]]) -> None:
        """
        Manually set a property for the object.
        """

        if not isinstance(combined_string, str):
            iterator = None
            try:
                iterator = iter(combined_string)
            except TypeError as this_exception:
                raise ValueError(
                    "Manual property form must either be a string or an iterable of strings."
                ) from this_exception
            for i in iterator:
                self.set_manual_property(i)
            return

        ApplicationProperties.verify_manual_property_form(combined_string)
        equals_index = combined_string.find(ApplicationProperties.__assignment_operator)
        property_key = combined_string[:equals_index].lower()
        property_value = combined_string[equals_index + 1 :]
        composed_property_value: Any = property_value

        if (
            property_value.startswith(
                ApplicationProperties.__manual_property_type_prefix
            )
            and len(property_value) >= 2
        ):
            composed_property_value = self.__adjust_property_type(property_value)

        # This is a bit of a kludge, but it works consistently.  The manually set property
        # is always a string.  If the string has no type information associatede with it,
        # it is eligible for conversion into one of the other types.  To properly denote that
        # the value did not have any type information, it is saved again in the dictionary
        # with a prefix of the separator character, to denote eligibility.
        else:
            self.__flat_property_map[
                f"{ApplicationProperties.__separator}{property_key}"
            ] = property_value

        self.__flat_property_map[property_key] = copy.deepcopy(composed_property_value)
        LOGGER.debug(
            "Adding configuration '%s' : {%s}",
            property_key,
            str(composed_property_value),
        )

    def __adjust_property_type(self, property_value: str) -> Any:
        composed_property_value: Any = None
        if property_value[1] == ApplicationProperties.__manual_property_type_string:
            composed_property_value = property_value[2:]
        elif property_value[1] == ApplicationProperties.__manual_property_type_integer:
            try:
                composed_property_value = int(property_value[2:])
            except ValueError as this_exception:
                raise ValueError(
                    f"Manual property value '{property_value}' cannot be translated into an integer."
                ) from this_exception
        elif property_value[1] == ApplicationProperties.__manual_property_type_boolean:
            composed_property_value = property_value[2:].lower() == "true"
        else:
            composed_property_value = property_value[1:]
        return composed_property_value

    # pylint: disable=too-many-arguments
    def get_property(
        self,
        property_name: str,
        property_type: type,
        default_value: Any = None,
        valid_value_fn: Optional[Callable[[Any], Any]] = None,
        is_required: bool = False,
        strict_mode: Optional[Any] = None,
    ) -> Any:
        """
        Get an property of a generic type from the configuration.
        """

        if strict_mode is None:
            strict_mode = self.__strict_mode

        if not isinstance(property_name, str):
            raise ValueError("The propertyName argument must be a string.")
        ApplicationProperties.verify_full_key_form(property_name)
        if not isinstance(property_type, type):
            raise ValueError(
                f"The property_type argument for '{property_name}' must be a type."
            )

        if default_value is not None:
            is_eligible = isinstance(default_value, property_type)
            if is_eligible and property_type == int and isinstance(default_value, bool):
                is_eligible = False

            if not is_eligible:
                raise ValueError(
                    f"The default value for property '{property_name}' must "
                    + f"either be None or a '{property_type.__name__}' value."
                )

        property_value = default_value
        property_name = property_name.lower()
        LOGGER.debug("property_name=%s", property_name)
        if property_name in self.__flat_property_map:
            property_value = self.__get_present_property(
                property_name,
                property_value,
                property_type,
                strict_mode,
                valid_value_fn,
            )
        elif is_required:
            raise ValueError(
                f"A value for property '{property_name}' must be provided."
            )
        return property_value

    # pylint: enable=too-many-arguments

    def __get_present_property_value(
        self, property_name: str, property_type: type
    ) -> Tuple[bool, Any]:
        found_value = self.__flat_property_map[property_name]
        is_eligible = isinstance(found_value, property_type)
        if is_eligible and property_type == int and isinstance(found_value, bool):
            is_eligible = False

        covertable_property_name = f"{ApplicationProperties.__separator}{property_name}"
        if (
            not is_eligible
            and property_type != str
            and self.__convert_untyped_if_possible
            and covertable_property_name in self.__flat_property_map
        ):
            found_value = self.__flat_property_map[covertable_property_name]
            # print(f"::{covertable_property_name}::{found_value}::")
            if property_type == bool:
                found_value = f"{ApplicationProperties.__manual_property_type_prefix}{ApplicationProperties.__manual_property_type_boolean}{found_value}"
            else:
                found_value = f"{ApplicationProperties.__manual_property_type_prefix}{ApplicationProperties.__manual_property_type_integer}{found_value}"
            with contextlib.suppress(ValueError):
                found_value = self.__adjust_property_type(found_value)
                is_eligible = True
            # print(f"::{covertable_property_name}::{found_value}::")
        return is_eligible, found_value

    # pylint: disable=too-many-arguments, broad-exception-caught
    def __get_present_property(
        self,
        property_name: str,
        property_value: Any,
        property_type: type,
        strict_mode: bool,
        valid_value_fn: Optional[Callable[[Any], Any]],
    ) -> Any:
        is_eligible, found_value = self.__get_present_property_value(
            property_name, property_type
        )
        if not is_eligible and strict_mode:
            raise ValueError(
                f"The value for property '{property_name}' must be of type '{property_type.__name__}'."
            )
        if is_eligible and valid_value_fn:
            try:
                valid_value_fn(found_value)
            except Exception as this_exception:
                is_eligible = False
                if strict_mode:
                    raise ValueError(
                        f"The value for property '{property_name}' is not valid: {str(this_exception)}"
                    ) from this_exception
        if is_eligible:
            property_value = found_value
        return property_value

    # pylint: enable=too-many-arguments, broad-exception-caught

    def get_boolean_property(
        self,
        property_name: str,
        default_value: Optional[bool] = None,
        is_required: bool = False,
        strict_mode: Optional[bool] = None,
    ) -> Optional[bool]:
        """
        Get a boolean property from the configuration.
        """
        return cast(
            bool,
            self.get_property(
                property_name,
                bool,
                default_value=default_value,
                valid_value_fn=None,
                is_required=is_required,
                strict_mode=strict_mode,
            ),
        )

    # pylint: disable=too-many-arguments
    def get_integer_property(
        self,
        property_name: str,
        default_value: Optional[int] = None,
        valid_value_fn: Optional[Callable[[int], Any]] = None,
        is_required: bool = False,
        strict_mode: Optional[bool] = None,
    ) -> Optional[int]:
        """
        Get an integer property from the configuration.
        """
        return cast(
            int,
            self.get_property(
                property_name,
                int,
                default_value=default_value,
                valid_value_fn=valid_value_fn,
                is_required=is_required,
                strict_mode=strict_mode,
            ),
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def get_string_property(
        self,
        property_name: str,
        default_value: Optional[str] = None,
        valid_value_fn: Optional[Callable[[str], Any]] = None,
        is_required: bool = False,
        strict_mode: Optional[bool] = None,
    ) -> Optional[str]:
        """
        Get a string property from the configuration.
        """
        return cast(
            str,
            self.get_property(
                property_name,
                str,
                default_value=default_value,
                valid_value_fn=valid_value_fn,
                is_required=is_required,
                strict_mode=strict_mode,
            ),
        )

    # pylint: enable=too-many-arguments

    def property_names_under(self, key_name: str) -> List[str]:
        """
        List of each of the properties in the map under the specified key.
        """
        ApplicationProperties.verify_full_key_form(key_name)
        return [
            next_key_name
            for next_key_name in self.__flat_property_map
            if next_key_name.startswith(key_name)
        ]

    # pylint: disable=too-many-boolean-expressions
    def __scan_map(
        self,
        config_map: Dict[Any, Any],
        current_prefix: str,
        allow_periods_in_keys: bool,
    ) -> None:
        for next_key, next_value in config_map.items():
            if not isinstance(next_key, str):
                raise ValueError(
                    f"All keys in the main dictionary and nested dictionaries must be strings (not `{next_key}`)."
                )
            if (
                " " in next_key
                or "\t" in next_key
                or "\n" in next_key
                or ApplicationProperties.__assignment_operator in next_key
                or (
                    ApplicationProperties.__separator in next_key
                    and not allow_periods_in_keys
                )
            ):
                raise ValueError(
                    f"Key string `{next_key}` cannot contain a whitespace character, "
                    + f"a '{ApplicationProperties.__assignment_operator}' character, or "
                    + f"a '{ApplicationProperties.__separator}' character."
                )
            if ApplicationProperties.__separator in next_key:
                next_key = f"'{next_key}'"

            if isinstance(next_value, dict):
                self.__scan_map(
                    next_value,
                    f"{current_prefix}{next_key}{self.__separator}",
                    allow_periods_in_keys,
                )
            else:
                new_key = f"{current_prefix}{next_key}".lower()
                self.__flat_property_map[new_key] = copy.deepcopy(next_value)
                LOGGER.debug(
                    "Adding configuration '%s' : {%s}", new_key, str(next_value)
                )

    # pylint: enable=too-many-boolean-expressions



"""
Module to provide for a manner to load an ApplicationProperties object from a JSON file.
"""

import json
import os
from typing import Any, Callable, Dict, Optional, Tuple

import pyjson5

from application_properties.application_properties import ApplicationProperties
from application_properties.application_properties_loader_helper import (
    ApplicationPropertiesLoaderHelper,
)


# pylint: disable=too-few-public-methods
class ApplicationPropertiesJsonLoader:
    """
    Class to provide for a manner to load an ApplicationProperties object from a JSON file.
    """

    # pylint: disable=too-many-arguments

    @staticmethod
    def load_and_set(
        properties_object: ApplicationProperties,
        configuration_file: str,
        handle_error_fn: Optional[Callable[[str, Optional[Exception]], None]] = None,
        clear_property_map: bool = True,
        check_for_file_presence: bool = True,
        load_as_json5_file: bool = False,
    ) -> Tuple[bool, bool]:
        """
        Load the specified file and set it into the given properties object.
        """
        if check_for_file_presence and (
            not os.path.exists(configuration_file)
            or not os.path.isfile(configuration_file)
        ):
            return False, False

        configuration_map: Dict[Any, Any] = {}
        handle_error_fn = (
            ApplicationPropertiesLoaderHelper.set_error_handler_if_not_set(
                handle_error_fn
            )
        )

        if load_as_json5_file:
            did_have_one_error, configuration_map = (
                ApplicationPropertiesJsonLoader.__load_as_json5(
                    configuration_file, handle_error_fn
                )
            )
        else:
            did_have_one_error, configuration_map = (
                ApplicationPropertiesJsonLoader.__load_as_json(
                    configuration_file, handle_error_fn
                )
            )

        did_apply_map = False
        if not did_have_one_error and configuration_map:
            try:
                properties_object.load_from_dict(
                    configuration_map,
                    clear_map=clear_property_map,
                    allow_periods_in_keys=True,
                )
                did_apply_map = True
            except ValueError as this_exception:
                formatted_error = (
                    f"Specified configuration file '{configuration_file}' "
                    + f"is not valid: {str(this_exception)}"
                )
                did_have_one_error = True
                handle_error_fn(formatted_error, this_exception)
        return did_apply_map and not did_have_one_error, did_have_one_error

    # pylint: enable=too-many-arguments

    @staticmethod
    def __load_as_json(
        configuration_file: str,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> Tuple[bool, Dict[str, Any]]:
        did_have_one_error = False
        configuration_map: Dict[Any, Any] = {}
        try:
            with open(configuration_file, encoding="utf-8") as infile:
                sample_line = infile.readline()
                cur_pos = infile.tell()
                infile.seek(0, os.SEEK_END)
                end_pos = infile.tell()
                if sample_line.strip() == "" and cur_pos == end_pos:
                    configuration_map = {}
                else:
                    infile.seek(0, os.SEEK_SET)
                    configuration_map = json.load(infile)
        except json.decoder.JSONDecodeError as this_exception:
            formatted_error = (
                f"Specified configuration file '{configuration_file}' "
                + f"is not a valid JSON file: {str(this_exception)}."
            )
            did_have_one_error = True
            handle_error_fn(formatted_error, this_exception)
        except OSError as this_exception:
            formatted_error = (
                f"Specified configuration file '{configuration_file}' "
                + f"was not loaded: {str(this_exception)}."
            )
            did_have_one_error = True
            handle_error_fn(formatted_error, this_exception)
        return did_have_one_error, configuration_map

    # pylint: disable=no-member
    @staticmethod
    def __load_as_json5(
        configuration_file: str,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> Tuple[bool, Dict[str, Any]]:
        did_have_one_error = False
        configuration_map: Dict[Any, Any] = {}

        try:
            with open(configuration_file, encoding="utf-8") as infile:
                configuration_map = pyjson5.load(infile)
        except pyjson5.Json5EOF:
            configuration_map = {}
        except pyjson5.Json5DecoderException as this_exception:
            formatted_error = (
                f"Specified configuration file '{configuration_file}' "
                + f"is not a valid JSON file: {str(this_exception)}."
            )
            did_have_one_error = True
            handle_error_fn(formatted_error, this_exception)
        except OSError as this_exception:
            formatted_error = (
                f"Specified configuration file '{configuration_file}' "
                + f"was not loaded: {str(this_exception)}."
            )
            did_have_one_error = True
            handle_error_fn(formatted_error, this_exception)
        return did_have_one_error, configuration_map

    # pylint: enable=no-member


# pylint: enable=too-few-public-methods



"""
Module to provide for a manner to load an ApplicationProperties object from a ini-type config file.
"""

import configparser
import os
from typing import Callable, Optional, Set, Tuple

from application_properties.application_properties import ApplicationProperties
from application_properties.application_properties_loader_helper import (
    ApplicationPropertiesLoaderHelper,
)


# pylint: disable=too-few-public-methods
class ApplicationPropertiesConfigLoader:
    """
    Class to provide for a manner to load an ApplicationProperties object from a ini-type config file.
    """

    # pylint: disable=too-many-arguments
    @staticmethod
    def load_and_set(
        properties_object: ApplicationProperties,
        configuration_file: str,
        section_header: Optional[str] = None,
        handle_error_fn: Optional[Callable[[str, Optional[Exception]], None]] = None,
        clear_property_map: bool = True,
        check_for_file_presence: bool = True,
    ) -> Tuple[bool, bool]:
        """
        Load the specified file and set it into the given properties object.
        """
        handle_error_fn = (
            ApplicationPropertiesLoaderHelper.set_error_handler_if_not_set(
                handle_error_fn
            )
        )
        (
            did_succeed,
            did_have_one_error,
        ) = ApplicationPropertiesConfigLoader.__check_for_file(
            configuration_file, check_for_file_presence, handle_error_fn
        )
        if not did_succeed:
            return did_succeed, did_have_one_error

        if clear_property_map:
            properties_object.clear()

        config_parser = ApplicationPropertiesConfigLoader.__read_configuration(
            configuration_file, handle_error_fn
        )
        if not config_parser:
            return False, True

        did_apply_one = False
        set_property_names: Set[str] = set()
        for next_section_name in config_parser.sections():
            (
                did_apply_one,
                did_have_one_error,
            ) = ApplicationPropertiesConfigLoader.__next_section(
                properties_object,
                next_section_name,
                configuration_file,
                handle_error_fn,
                section_header,
                config_parser,
                set_property_names,
                did_have_one_error,
                did_apply_one,
            )
        return did_apply_one and not did_have_one_error, did_have_one_error

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __next_section(
        properties_object: ApplicationProperties,
        next_section_name: str,
        configuration_file: str,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
        section_header: Optional[str],
        config_parser: configparser.ConfigParser,
        set_property_names: Set[str],
        did_have_one_error: bool,
        did_apply_one: bool,
    ) -> Tuple[bool, bool]:
        if not ApplicationPropertiesConfigLoader.__verify_section_name(
            properties_object, next_section_name, configuration_file, handle_error_fn
        ):
            did_have_one_error = True
        elif not (section_header and next_section_name != section_header):
            for item_pair in config_parser.items(next_section_name):
                if ApplicationPropertiesConfigLoader.__set_item(
                    properties_object,
                    set_property_names,
                    item_pair,
                    configuration_file,
                    handle_error_fn,
                    section_header,
                    next_section_name,
                ):
                    did_apply_one = True
                else:
                    did_have_one_error = True
                    break
        return did_apply_one, did_have_one_error

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __set_item(
        properties_object: ApplicationProperties,
        set_property_names: Set[str],
        item_pair: Tuple[str, str],
        configuration_file: str,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
        section_header: Optional[str],
        next_section_name: str,
    ) -> bool:
        item_name = item_pair[0]
        item_value = item_pair[1]
        try:
            properties_object.verify_full_key_form(item_name, "Configuration item name")
        except ValueError as this_exception:
            formatted_error = (
                f"Configuration item name '{item_name}' in file '{configuration_file}' "
                + f"is not a valid section name: {str(this_exception)}"
            )
            handle_error_fn(formatted_error, this_exception)
            return False

        full_property_name = (
            f"{item_name}" if section_header else f"{next_section_name}.{item_name}"
        )
        if not item_value.strip():
            formatted_error = f"Full configuration item name '{full_property_name}' in file '{configuration_file}' does not have a value assigned to it."
            handle_error_fn(formatted_error, None)
            return False

        if full_property_name in set_property_names:
            formatted_error = f"Full configuration item name '{full_property_name}' in file '{configuration_file}' occurs multiple times using different formats."
            handle_error_fn(formatted_error, None)
            return False

        full_property = f"{full_property_name}={item_value}"
        properties_object.set_manual_property(full_property)
        set_property_names.add(full_property_name)
        return True

    # pylint: enable=too-many-arguments

    @staticmethod
    def __read_configuration(
        configuration_file: str,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> Optional[configparser.ConfigParser]:
        config_parser = configparser.ConfigParser(allow_no_value=False)
        try:
            config_parser.read(configuration_file)
            return config_parser
        except configparser.Error as this_exception:
            formatted_error = (
                f"Specified configuration file '{configuration_file}' "
                + f"is not a valid config file: {str(this_exception)}."
            )
            handle_error_fn(formatted_error, this_exception)
            return None

    @staticmethod
    def __check_for_file(
        configuration_file: str,
        check_for_file_presence: bool,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> Tuple[bool, bool]:
        if not os.path.exists(configuration_file) or not os.path.isfile(
            configuration_file
        ):
            if check_for_file_presence:
                return False, False

            formatted_error = (
                f"Specified configuration file '{configuration_file}' does not exist."
            )
            handle_error_fn(formatted_error, None)
            return False, True
        return True, False

    @staticmethod
    def __verify_section_name(
        properties_object: ApplicationProperties,
        next_section_name: str,
        configuration_file: str,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> bool:
        try:
            properties_object.verify_full_key_form(
                next_section_name, "Configuration section name"
            )
        except ValueError as this_exception:
            formatted_error = (
                f"Configuration section name '{next_section_name}' in file '{configuration_file}' "
                + f"is not a valid section name: {str(this_exception)}"
            )
            handle_error_fn(formatted_error, this_exception)
            return False
        return True


# pylint: enable=too-few-public-methods



"""
Module to provide for a manner to load an ApplicationProperties object from a YAML file.
"""

import os
from typing import Any, Callable, Dict, Optional, Tuple

import yaml

from application_properties.application_properties import ApplicationProperties
from application_properties.application_properties_loader_helper import (
    ApplicationPropertiesLoaderHelper,
)


# pylint: disable=too-few-public-methods
class ApplicationPropertiesYamlLoader:
    """
    Class to provide for a manner to load an ApplicationProperties object from a YAML file.
    """

    # pylint: disable=too-many-arguments
    @staticmethod
    def load_and_set(
        properties_object: ApplicationProperties,
        configuration_file: str,
        section_header: Optional[str] = None,
        handle_error_fn: Optional[Callable[[str, Optional[Exception]], None]] = None,
        clear_property_map: bool = True,
        check_for_file_presence: bool = True,
    ) -> Tuple[bool, bool]:
        """
        Load the specified file and set it into the given properties object.
        """
        if check_for_file_presence and not (
            os.path.exists(configuration_file) and os.path.isfile(configuration_file)
        ):
            return False, False

        handle_error_fn = (
            ApplicationPropertiesLoaderHelper.set_error_handler_if_not_set(
                handle_error_fn
            )
        )

        (
            did_have_one_error,
            configuration_map,
        ) = ApplicationPropertiesYamlLoader.__load_yaml_file(
            configuration_file, handle_error_fn
        )

        did_apply_map = False
        if not did_have_one_error:
            if configuration_map and section_header:
                configuration_map = (
                    ApplicationPropertiesYamlLoader.__apply_section_header(
                        configuration_map, section_header
                    )
                )
            if configuration_map:
                try:
                    properties_object.load_from_dict(
                        configuration_map,
                        clear_map=clear_property_map,
                        allow_periods_in_keys=True,
                    )
                    did_apply_map = True
                except ValueError as this_exception:
                    formatted_error = (
                        f"Specified configuration file '{configuration_file}' "
                        + f"contains invalidly formatted data: {str(this_exception)}"
                    )
                    did_have_one_error = True
                    handle_error_fn(formatted_error, this_exception)
        return did_apply_map and not did_have_one_error, did_have_one_error

    # pylint: enable=too-many-arguments
    @staticmethod
    def __load_yaml_file(
        configuration_file: str,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> Tuple[bool, Optional[Dict[str, Any]]]:
        configuration_map: Optional[Dict[str, Any]] = {}
        did_have_one_error = False
        try:
            with open(configuration_file, "rb") as infile:
                configuration_map = yaml.safe_load(infile)
        except yaml.MarkedYAMLError as this_exception:
            formatted_error = (
                f"Specified configuration file '{configuration_file}' "
                + f"is not a valid YAML file: {str(this_exception)}."
            )
            did_have_one_error = True
            handle_error_fn(formatted_error, this_exception)
        except IOError as this_exception:
            formatted_error = (
                f"Specified configuration file '{configuration_file}' "
                + f"was not loaded: {str(this_exception)}."
            )
            did_have_one_error = True
            handle_error_fn(formatted_error, this_exception)

        if configuration_map is None:
            configuration_map = {}
        elif not isinstance(configuration_map, dict):
            formatted_error = f"Specified configuration file '{configuration_file}' is not a valid YAML file."
            did_have_one_error = True
            handle_error_fn(formatted_error, None)
        return did_have_one_error, configuration_map

    @staticmethod
    def __apply_section_header(
        configuration_map: Dict[str, Any], section_header: str
    ) -> Optional[Dict[str, Any]]:
        if section_header not in configuration_map:
            return None
        configuration_map = configuration_map[section_header]
        return configuration_map


# pylint: enable=too-few-public-methods
